/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.3 Automatisches Lernen
 * Ganglinien
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllernen.gllernen;

import java.util.Calendar;

import com.bitctrl.util.AbstractCronJob;
import com.bitctrl.util.Interval;
import com.bitctrl.util.Timestamp;

import de.bsvrz.iav.gllib.gllib.modell.parameter.PdGanglinienModellAutomatischesLernen;
import de.bsvrz.sys.funclib.bitctrl.daf.BetriebsmeldungDaten;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.bitctrl.modell.ObjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Führt einen Zyklus des Ganglinienlernens aus. Es wird für jeden
 * Messquerschnitt das Lernen mit den gleichen (aktuellen) Lernenparametern
 * durchgeführt.
 *
 * @author BitCtrl Systems GmbH, Schumann
 */
public class LernTask extends AbstractCronJob {

	/** Der Logger der Klasse. */
	private final Debug log = Debug.getLogger();

	/** Der Lernparameter mit dem gelernt werden soll. */
	private PdGanglinienModellAutomatischesLernen.Daten lernParam;

	/**
	 * Initialisiert den Lernprozess.
	 *
	 * @param lernParam
	 *            der Lernparameter.
	 */
	public LernTask(
			final PdGanglinienModellAutomatischesLernen.Daten lernParam) {
		super(lernParam.getAktualisierungsintervall());

		this.lernParam = lernParam;
		LogTools.log(log, GlLernenMsg.ConfigLernenInitialisiert,
				Verwaltung.getInstanz().getSemaphore().availablePermits());
	}

	/**
	 * Startet den Lernprozess für die angegebenen Messquerschnitte.
	 */
	@Override
	public synchronized void run() {
		PdGanglinienModellAutomatischesLernen.Daten param;
		final Interval lernzeitraum;
		final Calendar cal;
		final long aktuelleZeit;
		final long startLernzeitraum, endeLernzeitraum;

		cal = Calendar.getInstance();

		// Lernparameter zwischenspeichern, damit in diesem Zyklus für alle MQs
		// mit den selben Parameterdaten gelernt wird.
		param = getLernParam();

		// Lernzeitraum bestimmen
		aktuelleZeit = ObjektFactory.getInstanz().getVerbindung().getTime();
		cal.setTimeInMillis(aktuelleZeit);
		cal.set(Calendar.HOUR_OF_DAY, 24);
		cal.set(Calendar.MINUTE, 0);
		cal.set(Calendar.SECOND, 0);
		cal.set(Calendar.MILLISECOND, 0);
		startLernzeitraum = Verwaltung.getInstanz().getLetzteLernzeit();
		endeLernzeitraum = cal.getTimeInMillis() - param.getDatenMindestalter();
		LogTools.log(log, GlLernenMsg.FineInitialisiereLernen,
				Timestamp.absoluteTime(aktuelleZeit),
				Timestamp.absoluteTime(startLernzeitraum),
				Timestamp.absoluteTime(endeLernzeitraum));

		// Lernzeitraum validieren
		if ((endeLernzeitraum - startLernzeitraum) <= 0) {
			LogTools.log(log, GlLernenMsg.InfoLernenNichtNotwendig);
		} else {
			lernzeitraum = new Interval(startLernzeitraum, endeLernzeitraum);

			LogTools.log(log, GlLernenMsg.InfoLernenStart, "alle MQ", //$NON-NLS-1$
					lernzeitraum);

			// Starte für jeden zu lernenden MQ einen Thread so lange Threads
			// verfügbar sind, sonst warte bis wieder Theads frei sind.
			MessQuerschnittAllgemein[] messquerschnitte;

			messquerschnitte = Verwaltung.getInstanz().getMessquerschnitte()
					.toArray(new MessQuerschnittAllgemein[0]);
			for (int i = 0; i < messquerschnitte.length; ++i) {
				// for (final MessQuerschnittAllgemein mq :
				// Verwaltung.getInstanz().getMessquerschnitte()) {
				final Thread thread;
				final LernenMQ lernenMq;
				final MessQuerschnittAllgemein mq;

				mq = messquerschnitte[i];
				try {
					Verwaltung.getInstanz().getSemaphore().acquire();
				} catch (final InterruptedException ex) {
					LogTools.log(log, GlLernenMsg.WarningInterrupt, "alle MQ"); //$NON-NLS-1$
					break;
				}

				LogTools.log(log,
						new BetriebsmeldungDaten(mq.getSystemObject()),
						GlLernenMsg.InfoLernenStartMQ, mq, (i + 1),
						messquerschnitte.length);

				lernenMq = new LernenMQ(mq, lernzeitraum,
						param.getMaxVergleichsAbstand());
				thread = new Thread(lernenMq);
				thread.setName(GlLernenMsg.ApplikationsnameKurz + " : " + mq); //$NON-NLS-1$
				thread.start();
			}

			// Letzten Zeitpunkt des Lernens aktualisieren
			Verwaltung.getInstanz().setLetzteLernzeit(endeLernzeitraum);
			LogTools.log(log, GlLernenMsg.InfoLernenErfolgreich, "alle MQ"); //$NON-NLS-1$
		}
	}

	/**
	 * Gibt den aktuellen Parameter des automatischen Ganglinienlernens zurück.
	 *
	 * @return der Lernparameter
	 */
	public PdGanglinienModellAutomatischesLernen.Daten getLernParam() {
		return lernParam;
	}

	/**
	 * Legt den Parameter des automatischen Ganglinienlernens fest.
	 *
	 * @param lernParam
	 *            der neue Lernparameter.
	 */
	public void setLernParam(
			final PdGanglinienModellAutomatischesLernen.Daten lernParam) {
		this.lernParam = lernParam;
		setPattern(lernParam.getAktualisierungsintervall());
	}

}
