/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.3 Automatisches Lernen
 * Ganglinien
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllernen.gllernen;

import java.util.Calendar;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.Semaphore;

import com.bitctrl.Constants;
import com.bitctrl.i18n.Messages;

import de.bsvrz.iav.gllib.gllib.modell.parameter.PdGanglinienModellAutomatischesLernen;
import de.bsvrz.sys.funclib.bitctrl.daf.CronSchedulerDav;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensatzUpdateEvent;
import de.bsvrz.sys.funclib.bitctrl.modell.DatensatzUpdateListener;
import de.bsvrz.sys.funclib.bitctrl.modell.ObjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Verwaltet das Lernen und führt es zyklisch aus. Ist verantworlich für das
 * Überwachen der Lernparameter.
 *
 * @author BitCtrl Systems GmbH, Schumann
 */
public final class Verwaltung implements DatensatzUpdateListener {

	/** Das Singleton der Verwaltung. */
	private static Verwaltung singleton;

	/** Die Menge der Messquerschnitte für die gelernt wird. */
	private final Set<MessQuerschnittAllgemein> messquerschnitte = new HashSet<>();

	/** Der Logger der Klasse. */
	private final Debug log = Debug.getLogger();

	/** Der Scheduler der das Lernen zyklisch anstößt. */
	private final CronSchedulerDav scheduler;

	/** Die ID des Cron-Jobs für das Lernen. */
	private long jobId = -1;

	/** Der Zeitpunkt an dem das letzte Mal gelernt wurde. */
	private long letzteLernzeit;

	/** Verwaltet die maximale Anzahl gleichzeitiger Lernthreads. */
	private final Semaphore semaphore;

	/**
	 * Gibt die einzige Instanz der Verwaltung zurück.
	 *
	 * @return die Verwaltung.
	 */
	public static Verwaltung getInstanz() {
		if (singleton == null) {
			singleton = new Verwaltung();
		}
		return singleton;
	}

	/**
	 * Initialisiert die Verwaltung. Die letzte Lernzeit wird mit heutiger Tag
	 * minus Datenmindestalter initialisiert und der Scheduler für das Lernen
	 * gestartet.
	 */
	private Verwaltung() {
		final PdGanglinienModellAutomatischesLernen param;
		final PdGanglinienModellAutomatischesLernen.Daten datum;
		final ObjektFactory factory;
		final Calendar cal;
		final int slots;

		factory = ObjektFactory.getInstanz();

		// Anzahl der maximal gleichzeitig aktiven Lernthreads festlegen
		if (GlLernen.getMaxThreads() > 0) {
			// Anzahl durch Kommandozeilenargument festgelegt
			slots = GlLernen.getMaxThreads();
		} else if (System.getenv("NUMBER_OF_PROCESSORS") != null) { //$NON-NLS-1$
			// Standard: doppelte Anzahl der vorhandenen (C)PUs
			slots = Integer.parseInt(System.getenv("NUMBER_OF_PROCESSORS")) * 2; //$NON-NLS-1$
		} else {
			// Keine systemabhängige Festlegung möglich
			slots = 2;
		}
		semaphore = new Semaphore(slots);

		// Parameterdaten initial auslesen
		param = factory.getAOE().getParameterDatensatz(
				PdGanglinienModellAutomatischesLernen.class);
		datum = param.abrufenDatum();
		if (!datum.isValid() || (datum.getAktualisierungsintervall() == null)) {
			throw new IllegalStateException(Messages.get(
					GlLernenMsg.IllegalStateExceptionLernParameterUngueltig,
					param, datum));
		}

		// Letzte Lernzeit initialisieren: heute - Datenmindestalter
		cal = Calendar.getInstance();
		cal.setTimeInMillis(factory.getVerbindung().getTime());
		cal.set(Calendar.HOUR_OF_DAY, 24);
		cal.set(Calendar.MINUTE, 0);
		cal.set(Calendar.SECOND, 0);
		cal.set(Calendar.MILLISECOND, 0);
		cal.add(Calendar.DAY_OF_YEAR, (int) -(datum.getDatenMindestalter()
				/ Constants.MILLIS_PER_DAY));
		setLetzteLernzeit(cal.getTimeInMillis());

		// Scheduler starten
		scheduler = new CronSchedulerDav();
		scheduler.start();

		// Auf Parameteränderungen anmelden
		param.addUpdateListener(this);

		LogTools.log(log, GlLernenMsg.ConfigVerwaltungBereit);
	}

	/**
	 * Legt die Menge der Messquerschnitt für die gelernt werden soll fest.
	 *
	 * @param messquerschnitte
	 *            eine Menge von Messquerschnitten.
	 */
	public void setMessquerschnitte(
			final Collection<MessQuerschnittAllgemein> messquerschnitte) {
		this.messquerschnitte.clear();
		this.messquerschnitte.addAll(messquerschnitte);
	}

	/**
	 * Gibt die Menge der Messquerschnitte zurück, für die gelernt wird.
	 *
	 * @return die Messquerschnitte.
	 */
	public Set<MessQuerschnittAllgemein> getMessquerschnitte() {
		return messquerschnitte;
	}

	/**
	 * Verarbeitet empfangene Parameteränderungen und aktualisiert das Intervall
	 * des Scheduler.
	 */
	@Override
	public void datensatzAktualisiert(final DatensatzUpdateEvent event) {
		final PdGanglinienModellAutomatischesLernen.Daten datum;
		final LernTask lernTask;

		datum = (PdGanglinienModellAutomatischesLernen.Daten) event.getDatum();
		if (datum.isValid() && (datum.getAktualisierungsintervall() != null)) {
			LogTools.log(log, GlLernenMsg.InfoParameterAenderung, datum);
			if (jobId < 0) {
				// Parameter wird zum ersten Mal ausgewertet, ergo Job anlegen.
				jobId = scheduler.schedule(new LernTask(datum));
			}
			if (!scheduler.isStarted()) {
				LogTools.log(log, GlLernenMsg.InfoLernenStartetWieder, datum);
				scheduler.start();
			}
		} else {
			LogTools.log(log, GlLernenMsg.ErrorUngueltigerLernParameter, datum);
			scheduler.stop();
			return;
		}

		lernTask = (LernTask) scheduler.getCronJob(jobId);
		lernTask.setLernParam(datum);
	}

	/**
	 * Gibt den Zeitpunkt an dem das letzte Mal gelernt wurde zurück.
	 *
	 * @return der letzte Lernzeitpunkt.
	 */
	public long getLetzteLernzeit() {
		return letzteLernzeit;
	}

	/**
	 * Legt den Zeitpunkt fest, an dem das letzte Mal gelernt wurde.
	 *
	 * @param letzteLernzeit
	 *            der letzte Lernzeitpunkt.
	 */
	public void setLetzteLernzeit(final long letzteLernzeit) {
		LogTools.log(log, GlLernenMsg.FineNeueLetzteLernzeit, letzteLernzeit);
		this.letzteLernzeit = letzteLernzeit;
	}

	/**
	 * Gibt die Semaphore für Verwaltung der Lernthreads zurück.
	 *
	 * @return die Semaphore.
	 */
	public Semaphore getSemaphore() {
		return semaphore;
	}

}
